//Integración Romberg
clear
clc
//El algoritmo de la extrapolación repetida de Richardson 
//o integración Romberg, se basa en la utilización de evaluaciones de una 
//función para crear una regla de cuadratura utilizando un polinomio de grado superior.
//Por ejemplo, dado que conocemos que la regla del trapezoide es de O(h^2) en exactitud,
//podemos combinar una regla del trapezoide con una a dos veces la discretización 
//para obtener una regla de Simpson.
//De la misma manera, podemos combinar dos reglas de Simpson par obtener una nueva regla con 
//error de O(h^6), y así sucesivamente.
//
//El algoritmo es como sigue:
//T1
//T2  S2
//T4  S4  P4
//T8  S8  P8  Q8
//T16 S16 P16 Q16 R16
//y así sucesivamente.
//La idea es que las reglas del trapezoide T1 y T2 de uno y dos paneles se combinen para formar 
//la regla de Simpson de dos paneles, S2.
//De la misma manera, las reglas de Simpson de cuatro y ocho paneles se combinan para formar
//la regla de ocho paneles en el orden superior siguiente.
//Este algoritmo avanza tanto como querramos, sólo depende del número de evaluaciones que se 
//hayan efectuado de la función.
//
//Implementemos esta técnica. Iremos hasta el quinto nivel de iteración.
//Comenzamos construyendo las reglas trapezoidales. Operamos sobre la función exp(x) en el
//intervalo [-1,1]. Luego:
n=5; //Niveles de iteración
a=-1;
b=1;
h=(b-a)/2^n;
t(1,1)=(exp(a)+exp(b))/2;
for i=2:n+1,
 t(i,1)=t(i-1,1);
 for j=1:2^(i-2),
  t(i,1)=t(i,1)+exp(a+h*2^(n-i+1)+h*(j-1)*2^(n-i+2));
 end
end
for i=1:n+1,
 t(i,1)=t(i,1)*h*2^(n+1-i);
end
//
//Este algoritmo se construyó para la regla trapezoidal de manera de efectuar el mínimo
//número de llamados de la función. Recordemos que para 2^k paneles (con k=0,1,2 ...n),
//se requieren (2^k)+1 evaluaciones de la función; estas evaluaciones se vuelven a utilizar
//con 2^(k+1) paneles (en total 2^(k+1)+1 evaluaciones). Por consiguiente, en cada etapa sólo se
//necesita evaluar la función en 2^k nuevos puntos intermedios a los anteriores.
//Si estudiamos la estructura de la expresión de arriba, vemos que ésta se verifica.
//No se requieren evaluaciones adicionales de la función para la extrapolación sucesiva de
//Richardson
//
//Ahora que hemos construido las estimaciones de la regla trapezoidal, grafiquemos los errores
//de arriba:
figure(1);
plot2d('nl',abs(t(:,1)-(exp(1)-exp(-1))))
gca().grid=[1 1 1]
xlabel('Nivel de iteración')
ylabel('Error')
title('Regla del Trapecio')
set(gca(),"auto_clear","on")
//Como podemos ver, el error es recortado en un factor de cuatro cada vez que el intervalo es
//llevado a la mitad.
//
//Apliquemos ahora el método de extrapolación sucesiva de Richardson.
//Cada regla sucesiva será de orden h^2 más alta que aquella utilizada para calcularla. Por lo
//tanto podemos utilizar la fórmula de Simpson de 2n puntos a partir de la fórmula:
//S2n = (4*T2n - Tn)/(4-1) y la siguiente fórmula de orden superior:
//P2n = (16*S2n - Sn)/(16-1) y así sucesivamente:
for j=2:n+1
 for i=n+1:-1:j
  t(i,j)=(4^(j-1)*t(i,j-1)-t(i-1,j-1))/(4^(j-1)-1);
 end
end
figure(2);
plot2d('nl',abs(t(n+1,:)-(exp(1)-exp(-1))))
gca().grid=[1 1 1]
xlabel('Nivel de iteración')
ylabel('Error')
title('Integración Romberg')
set(gca(),"auto_clear","on")
//y esto es todo!. La matriz t es de la forma triangular inferior, y contiene la regla del
//trapezoide, la de Simpson y estimaciones de cuadratura de orden superior.
//Veamos esta matriz:
//
t
//
//Como puede apreciarse, las estimaciones rápidamente convergen al valor correcto.
//Resulta de interés graficar los errores de la última fila. Esta fila contiene las
//estimaciones de cuadratura de orden h^2k (k=1, 2, ..., (n+1)), siendo la primer columna el
//resultado de la regla del trapezoide, la segunda el resultado de la regla de Simpson y así
//sucesivamente. Por lo tanto:
//
figure(3);
plot2d('nl',[1:6]',[abs(t(n+1,:)-(exp(1)-exp(-1)))',abs(t(:,1)-(exp(1)-exp(-1)))],[1:2],leg='Romberg 32 paneles@Trapezoide')
set(gca(),"auto_clear","on")
gca().grid=[1 1 1]
xlabel('Nivel de iteración')
ylabel('Error')
title('Comparación de Errores de las Reglas de Cuadratura')
set(gca(),"auto_clear","on")
//Por consiguiente, después de la quinta iteración tenemos una regla que es del orden h^12,
//y el error es del orden del error de redondeo, al menos para esta función uniforme.
//No podemos ir mucho más allá, dado que estamos limitados por la exactitud en la doble
//precisión de la computadora. 
//En conclusión, la regla de extrapolación sucesiva de Richardson representa un enfoque simple
//y útil para efectuar estimaciones de cuadratura si no se tiene a mano una tabla de nodos y
//pesos de cuadratura Gaussiana. La diferencia entre los dos últimos elementos de la última fila 
//suministra una estimación del error de cuadratura. Un par G7K15 Gauss-Kronod podría alcanzar
//un más que elevado nivel de exactitud así como una estimación del error con un poco menos de
//la mitad de evaluaciones de la función utilizada aquí.
